/*++++++++++++++
.IDENTIFICATION readme2f.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    Unix, CDS Catalogues
.KEYWORDS       Table, FITS
.VERSION  1.0   05-Mar-1998
.VERSION  1.1   24-Apr-1998: Pb arrays...
.VERSION  1.2   09-Aug-2001: Remove initial _ in names (for g77)
.VERSION  1.3   28-Jul-2003: Ignore the NODB fields, as well as "---" fields
.VERSION  1.4   20-Feb-2004: Take care of arrays
.VERSION  1.5   11-Sep-2004: Problem with --- labels !!
.VERSION  1.6   19-Feb-2005: Bug fixed in var_list (reported by
.VERSION  1.7   02-Jun-2007: Cosmetics (-Wall)
.VERSION  1.8   22-Mar-2013: 64-bit
			 Boud Roukema <boud@astro.uni.torun.pl>)
.COMMENTS       Generate FORTRAN from ReadMe
---------------*/

#define VERSION "1.7"

#ifndef DEBUG
#define DEBUG 0
#endif
#ifndef TEST
#define TEST 0
#endif
/*static int vartest = TEST;*/

#include <catstd.h>
#include <filutil.h>
#include <string.h>	/* Standard String manipulations */
#include <ctype.h>

#define MAX(x,y)        ((x) >= (y) ? (x) : (y))
#define MIN(x,y)        ((x) >= (y) ? (y) : (x))
#define ITEMS(a)        sizeof(a)/sizeof(a[0])

static char verbop = DEBUG;	/* Verbose option       */
static char oall = 0 ;

static char *ddat;		/* Directory with data	 */
static char *intro;		/* Name of Description File */

#if DEBUG || TEST
static char inbyte;
#endif

typedef struct s_id { 
   struct s_id *next, *prev ; 
   short orilen; 
   short arraylen;		/* Added v1.4	*/
   char name[2]; 
} ID ;
static ID *last_id; 

typedef struct s_var {
    struct s_var *next, *prev ;
    char recname[20] ;
    int fno ;
    FILINFO *pf ;
    ID *coo, *pid ;		/* F77 Names 	*/
    char *fmt;			/* F77 Format	*/
    char *vlist;		/* List of vars	*/
} FILEVAR ;
static FILEVAR *last_var ;

#define PGM "readme2f_1.5"

static char usage[] = "\
Usage: readme2f [-v] [-a] [-d dir] [-i file] [file...]" ;

static char help[] = "\
       This program generates a FORTRAN template to read the data tables\n\
       described by a standard ReadMe description file. Available options:\n\
      -v: verbose option\n\
      -a: load data as arrays (the whole table is loaded into memory)\n\
 -i file: specifies the name of description file (default 'ReadMe')\n\
  -d dir: specifies the directory containing the data files (default cwd)\n\
    file: file concerned\n\
" ;

static char header[] = "\
*  This code was generated from the ReadMe file documenting a catalogue\n\
*  according to the \"Standard for Documentation of Astronomical Catalogues\"\n\
*  currently in use by the Astronomical Data Centers (CDS, ADC, A&A)\n\
*  (see full documentation at URL http://vizier.u-strasbg.fr/doc/catstd.htx)\n\
*  Please report problems or questions to   question@simbad.u-strasbg.fr\n\
" ;
static char explain_null[] = "\
*  Unspecified or NULL values, generally corresponding to blank columns,\n\
*  are assigned one of the following special values:\n\
*     rNULL__    for unknown or NULL floating-point values\n\
*     iNULL__    for unknown or NULL   integer      values\n\
"; 
  static char Crule[] = "\
C=============================================================================";
  static char crule[] = "\
c - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ";

/*==================================================================
		Internal Utilities
 *==================================================================*/
static char *afac(double v)
/*++++++++++++++++
.PURPOSE  Ensure the char representation of a number includes a .
.RETURNS  The number 
.REMARKS  From internal buffer
-----------------*/
{
  static char avalue[200] ;
  char *p ;
    sprintf(avalue, "%g", v) ;
    for (p=avalue; *p; p++) {
	if (*p == '.') return(avalue) ;
    }
    *(p++) = '.' ;
    *p = 0 ;
    return(avalue) ;
}

static ID *f77name(char *str)
/*++++++++++++++++
.PURPOSE  Transform a name to be FORTRAN-compatible
.RETURNS  The transformed name / NULL for names to ignore
.REMARKS  Newly allocated string.
-----------------*/
{
  static char name[33] ;
  ID *pid ;
  char c1, *p, *e ;
  int len, i ;

     /* Ignore names made of just - */
     if (*str == '-') {
	for (p=str; *p == '-'; p++) ;
	if (*p) fprintf(stderr, "****Column '%s' ignored (-)\n", str);
	name[0] = 0; len = 0;
	goto Install_Name;
	/* return((ID *)0); */
     }
     /* c1 = isalpha(*str) || (*str == '_') ? 0 : '_' ; Starting _ unaccepted */
     c1 = isalpha(*str) ? 0 : 'v' ;
     p = name ; e = name + sizeof(name)-4 ;
     if (c1) *(p++) = c1;
     for (; (p<e)&& *str; p++, str++) {
	 if (isalnum(*str)) *p = *str ;
	 else if ((*str == ')') && (!str[1])) *p = 0 ;
	 else *p = '_' ;
     }
     *p = 0 ; len = p-name ;

     /* Look: is the variable already used ? */
     for (pid=last_id; pid; pid = pid->prev) {
	if (len != pid->orilen) continue ;
	if (strncasecmp(pid->name, name, len) == 0) break ;
     }
     if (pid) {		/* Name already in use... */
	p = pid->name + pid->orilen ;
	if (*p) { if (!isdigit(*p)) p++; }
	i = atoi(p) + 1 ;
	p = name + strlen(name) ;
	sprintf(p, "_%d", i) ;
     }

     /* Allocate the new name */
  Install_Name:
     pid = (ID *)malloc(sizeof(ID) + strlen(name) + 1) ;
     pid->orilen = len ;
     pid->arraylen = 0 ;
     strcpy(pid->name, name) ;
     if (last_id) last_id->next = pid ;
     pid->prev = last_id ;
     pid->next = (ID *)0 ;
     last_id = pid ;

     return(pid) ;
}

static void prc(char *str)
/*++++++++++++++++
.PURPOSE  Write Continuations lines
.RETURNS  ---
.REMARKS  Last line NOT terminated by newline
-----------------*/
{
  char *p, *pv ;
  int len, len1 ;
    for (p=str; *p; p += len) {
	putchar('\n') ;
	printf("     +  ") ;
	len = len1 = 0 ; pv = p ;
	while ((*pv) && (len < 62)) {
	    len1 = strloc(pv, ',') ; ;
	    if (pv[len1]) len1++ ;
	    len += len1 ;
	    pv += len1 ;
	}
	if (len > 62) len -= len1 ;
	if (len == 0) len = len1 ;
	fwrite(p, 1, len, stdout) ;
    }
}

#if 0
static void prs(char *str)
/*++++++++++++++++
.PURPOSE  Write a a string, replacing non acceptables chars by _
.RETURNS  ---
.REMARKS  
-----------------*/
{
  char *p ;
    for (p=str; *p; p++) {
	if (isalnum(*p)) { putchar(*p); continue; }
	putchar('_') ;
    }
}
#endif

static void prt(char *str)
/*++++++++++++++++
.PURPOSE  Write Text (newlines --> comments)
.RETURNS  ---
.REMARKS  Last line NOT terminated by newline
-----------------*/
{
  char *p;

    for (p=str; *p ; p++) {
        putchar (*p) ;
        if (*p == '\n') putchar('*') ;
    }
}

static char *var_fmt(TABINFO *pt, ID *pid)
/*++++++++++++++++
.PURPOSE  Computes the Format
.RETURNS  Text of Format (no brackets)
.REMARKS  Newly allocated string
-----------------*/
{
  COLINFO *cc, *ce ;
  ID *pn ;
  int o ;
  char sep, *p, *list ;

    p = list = malloc(12*pt->nfmt) ;
    o = 0 ; sep = 0 ;
    pn = pid;
    for (cc=pt->afmt, ce=cc+pt->nfmt; cc<ce; cc++, pn=pn->next) {
	if(!pn->name[0]) continue;		/* Name was "---" */
	if (sep) *(p++) = sep ; sep = ',' ;
	if (o < cc->b0) sprintf(p, "%dX,", cc->b0 - o), p += strlen(p) ;
	o = cc->b1 + 1 ;
	if (cc->rep > 1) sprintf(p, "%d", cc->rep), p += strlen(p) ;
	sprintf(p, "%c%d", cc->ff, cc->len), p += strlen(p) ;
	if (cc->ff == 'I') continue ;
	if (cc->ff == 'A') continue ;
	sprintf(p, ".%d", cc->len-cc->odp-1) ;
	p += strlen(p) ;
    }
    *p = 0 ;
    return(list) ;
}

static void set_null(TABINFO *pt, ID *pid, char *index, char *recname)
/*++++++++++++++++
.PURPOSE  Replace the NULL values
.RETURNS  ---
.REMARKS  
-----------------*/
{
  COLINFO *cc, *ce ;
  ID *pn ;
  int8 *llim ; double *flim ;
  char fname[80], *name ;

    pn = pid ; 
    if (index) fname[0] = *index ; else fname[0] = 0 ;
    name = fname ;
    for (cc=pt->afmt, ce=cc+pt->nfmt; cc<ce; cc++, pn=pn->next) {
	if (cc->elims&F_REQD) continue ;
	if (cc->flags&F_NODB) continue;		/* Ignore this column */
	if(!pn->name[0]) continue;		/* Name was "---" */
	if (cc->ff == 'A') continue ;
	if (fname[0]) sprintf(fname, "%s(%s)", pn->name, index) ;
	else name = pn->name ;
	if (cc->elims&F_NULL) {		/* Explicit NULL */
	    if (cc->elims&F_DBLE) {
	        flim = (double *)cc->lims;
	        printf("        if (%s .EQ. %g) %s =  rNULL__\n", 
		    name, flim[2], name) ;
		continue ;
	    } 
	    llim = (int *)cc->lims;
#if int8l == 1
	    printf("        if (%s .EQ. %ld) %s =  iNULL__\n", 
		name, llim[2], name) ;
#else
	    printf("        if (%s .EQ. %lld) %s =  iNULL__\n", 
		name, llim[2], name) ;
#endif
	    continue ;
	}
	printf("        if(%s(%d:%d) .EQ. '') %s = ", recname,
	    cc->b0+1, cc->b1+1, name);
	putchar(cc->ff == 'I' ? 'i' : 'r') ;
	printf("NULL__\n") ;
    }

}

static void set_coo(TABINFO *pt, ID *pid, char *index, ID *coo)
/*++++++++++++++++
.PURPOSE  Compute RA + Dec in Degrees.
.RETURNS  ---
.REMARKS  
-----------------*/
{
static char com_fmt[] = "\
c       Derive coordinates %s and %s from input data\n\
c       (%s and %s are set to rNULL_ when unknown)\n\
" ;
  COLINFO *cc, *ce;
  ID *pn, *pra, *pde ;
  char pindex[80] ;
  double factor ;
  char *u, *signame ;
  int ira, ide ;
  char detype, commented ;

    pra = coo ; pde = coo->next ;

    if (index && *index) sprintf(pindex, "(%s)", index) ;
    else pindex[0] = 0 ;
    ira = ide = 0 ; 
    detype = commented = 0 ;
    signame = (char *)0 ;

    for (pn=pid, cc=pt->afmt, ce=cc+pt->nfmt; cc<ce; cc++, pn=pn->next) {
	if (!(cc->flags&F_MPOS)) continue;
	if (!commented) printf(com_fmt, pra->name, pde->name,
	    pra->name, pde->name) ;
	commented += 1 ;
	u = unit_si(cc->unit, &factor) ;
	switch(cc->label[0]) {
	  case 'R': 	/* Right Ascension */
	    if (ira == 0) printf("        %s%s = %s%s\n", pra->name, pindex,
		pn->name, pindex) ;
	    if (*u == 's') factor /= 86400. ;
	    factor *= 360. ;
	    printf("        if(%s%s .GE. 0) %s%s=", 
		pn->name, pindex, pra->name, pindex) ;
	    if (ira) printf("%s%s+", pra->name, pindex) ;
	    printf("%s%s", pn->name, pindex) ;
	    if (factor > 1.0000001) printf("*%s", afac(factor)) ;
	    else if (factor < 0.9999999) printf("/%s", afac(1./factor)) ;
	    putchar('\n') ;
	    ira++ ;
	    break ;
	  case_dec:
	  case 'D': 	/* Declination */
	    if (cc->ff == 'A') {	/* Sign */
		signame = pn->name ;
		continue ;
	    }
	    if (*u) continue ;		/* Can't be a declination */
	    if (ide == 0) printf("        %s%s = %s%s\n", pde->name, pindex,
		pn->name, pindex) ;
	    factor *= 360. ;
	    if (ide || (factor > 1.0000001) || (factor < 0.9999999)) {
	        printf("        if(%s%s .GE. 0) %s%s=", 
		    pn->name, pindex, pde->name, pindex) ;
	        if (ide) printf("%s%s+", pde->name, pindex) ;
	        printf("%s%s", pn->name, pindex) ;
	        if (factor > 1.0000001) printf("*%s", afac(factor)) ;
	        else if (factor < 0.9999999) printf("/%s", afac(1./factor)) ;
	        putchar('\n') ;
	    }
	    ide++ ;
	    break ;
	  case 'N': 	/* North Polar Distance */
	    detype = 'N' ;
	    goto case_dec ;
	  case 'S':	/* South Polar Distance */
	    detype = 'S' ;
	    goto case_dec ;
	  default:
	    if (verbop) fprintf(stderr, "++++Ignored compoment of MPOS: %s\n",
		cc->label) ;
	}
    }
    
    /* Change the sign if necessary */
    if (signame) printf("        if(%s%s.EQ.'-'.AND.%s%s.GE.0) %s%s=-%s%s\n", 
	signame,pindex, pde->name,pindex, pde->name,pindex, pde->name,pindex) ;
    if (detype) {
	if (detype == 'N') printf("        if(%s%s .GE. 0.)%s%s=90.-%s%s\n", 
	    pde->name,pindex, pde->name,pindex, pde->name,pindex) ;
	if (detype == 'S') printf("        if(%s%s .GE. 0.)%s%s=%s%s-90.0\n", 
	    pde->name,pindex, pde->name,pindex, pde->name,pindex) ;
    }
}

static char *var_list(ID *f77names, char *index)
/*++++++++++++++++
.PURPOSE  Write in a newly allocated string the list of variables
.RETURNS  Text of strings separated by commas
.REMARKS  
-----------------*/
{
  ID *pn ;
  int len, lindex, is_array ; 
  char sep, *p, *list ;

    lindex = 0 ;
    if (index && *index) lindex = 2 + strlen(index) ;
    /* Count the length required */
    for (len=0, pn=f77names; pn; pn=pn->next) {
	len += strlen(pn->name) + 1 /*,*/ + lindex ;
	if (pn->arraylen>1) len += 25;	/* ((j__),j__=1,...) */
    }
    p = list = malloc(len+1) ;

    /* Append the names */
    for (sep=0, pn=f77names; pn; pn=pn->next) {
	if (!pn->name[0])	/* Invalid name (---) */
	    continue;
	is_array = pn->arraylen>1;
	if(sep) *(p++) = sep; sep = ',' ;
	if (is_array) *(p++) = '('; 
	strcpy(p, pn->name) ; p += strlen(p) ;
	if (lindex || is_array) {
	    *(p++) = '(';
	    if (is_array) {
		strcpy(p, "j__"); p += strlen(p);
		if (index) *(p++) = ',';
	    }
	    if (index) strcpy(p, index), p += strlen(p) ;
	    if (is_array) {
		*(p++) = ')';
		sprintf(p, ",j__=1,%d", pn->arraylen);
		p += strlen(p);
	    }
	    *(p++) = ')';
	}
    }
    *p = 0 ;
    return(list) ;
}

static ID *dcl_coo(FILINFO *pf, char *dim)
/*++++++++++++++++
.PURPOSE  Find out the RA+DEc declarations for a table
.RETURNS  The list of FORTRAN names
.REMARKS  
-----------------*/
{
  TABINFO *pt ;
  COLINFO *cc;
  ID *f77names, *pid ;
  int i ; 
  char *p ;
    if (!pf->equinox) return((ID *)0) ;
    pt = pf->tabform ;
    if (pt->posfmt[0] < 0) return((ID *)0) ;
    i = pt->posfmt[2] ;
    if (i < 0) return((ID *)0) ;
    cc = pt->afmt + i;
    if (cc->label[0] == 'p') return((ID *)0) ;
    f77names = pid = f77name("RA_") ;
    printf("      real*%-8d %-11s", 8, pid->name) ; 
    if (dim) printf("(%s)", dim);
    printf(" ! (deg) Right Ascension %s", pf->equinox) ;
    if ((p = pf->epoch)) if (*p) printf(", Epoch=%s", p) ;
    putchar('\n') ;
    pid = f77name("Dec") ;
    printf("      real*%-8d %-11s", 8, pid->name) ; 
    if (dim) printf("(%s)", dim);
    printf(" ! (deg)     Declination %s", pf->equinox) ;
    if ((p = pf->epoch)) if (*p) printf(", Epoch=%s", p) ;
    putchar('\n') ;

    return(f77names) ;
}

static ID *dcl_f77(TABINFO *pt, char *dim)
/*++++++++++++++++
.PURPOSE  Edit the FORTRAN declarations
.RETURNS  The list of FORTRAN names
.REMARKS  
-----------------*/
{
  COLINFO *cc, *ce ;
  SECTION *s ;
  ID *f77names, *pid ;
  int len, i ; 
  char *p ;

    f77names = (ID *)0 ;

    for (cc=pt->afmt, ce=cc+pt->nfmt; cc<ce; cc++) {
	pid = f77name(cc->label) ;
	if (!pid) continue;			/* Column to ignore   */
	if (!f77names) f77names = pid ;
	if (!*pid->name) continue;		/* V1.5 name="---"    */
	pid->arraylen = cc->rep;		/* Added v1.4 */
	if (cc->flags&F_NODB) continue;		/* Ignore this column */
        printf("      ") ; len = 0 ;
	switch(cc->ff) {
	case 'A': printf("character*%-3d ", cc->len) ; break; 
	case 'I': printf("integer*%-5d ", 4) ; 	break; 
	case 'D': len = 8 ; break ;
	case 'E': len = 4 ; break ;
	case 'F': len = cc->len <= 6 ? 4 : 8 ; break ;
	default:  len = 8 ;
	}
	if (len) printf("real*%-8d ", len) ;
	printf("%-11s",pid->name) ;
	if (cc->rep > 1) {
	    printf("(%d", cc->rep) ;
	    if (dim) printf(",%s", dim) ;
	    putchar(')'); 
	}
	else if (dim) printf("(%s)", dim);
	printf(" ! ") ;
	p = cc->unit; len = 40 ;
	if (*p && (*p != '-')) {
	    printf("(%s) ", p); 
	    len -= (3+strlen(p)) ;
	}
	prt(cc->text) ;
	putchar('\n') ;
    }

    /* Write out all notes related to Table */
    s = pt->section ; s = s->next ;
    while (s->class == SEC_Note) {
	if (s->header) printf("*%s\n", s->header) ;
	if (s->ebody > s->abody) {
	    putchar('*') ;
	    p = s->ebody; i = *p ; *p = 0 ;
	    prt(s->abody); *p = i ;
	    putchar('\n') ;
	}
	s = s->next ;
    }
    return(f77names) ;
}

static void cad_f77(FILINFO *pf)
/*++++++++++++++++
.PURPOSE  Edit the program for 1 file
.RETURNS  ---
.REMARKS  All details saved in last_var
-----------------*/
{
  static int fno = 0 ;
  char parows[10], afno[10], *index;
  FILEVAR *fv ;
  ID *pid, *coo ;

    /* Define a name for record and number of records */
    if (fno) sprintf(afno, "%d", fno) ; else afno[0] = 0 ;
    index = oall ? "i__" : "" ;
    fno++ ;

    /* Allocate a FILEVAR, and link it */
    fv = (FILEVAR *)malloc(sizeof(FILEVAR));
    if (last_var) last_var->next = fv ;
    fv->prev = last_var ;
    fv->next = (FILEVAR *)0 ;
    fv->fno = fno ;
    last_var = fv ;

    fv->pf = pf ;
    sprintf(fv->recname, "ar__%s", afno) ;
    sprintf(parows,  "nr__%s", afno) ;
    printf("\n%s\n", crule) ;
    printf("\nC  Declarations for '%s'\t! %s\n\n", pf->dfile, pf->text) ;
    printf("      integer*4 %s\n", parows);
    printf("      parameter (%s=%d)\t! Number of records\n", 
	parows, pf->records) ;
    printf("      character*%d %-7s\t! Full-size record\n\n", pf->reclen, 
	fv->recname) ;

    coo = dcl_coo(pf, oall ? parows : 0) ;
    pid = dcl_f77(pf->tabform, oall ? parows : 0) ;
    fv->coo = coo ; fv->pid = pid ;
    fv->fmt   = var_fmt(pf->tabform, pid) ;
    fv->vlist = var_list(pid, index) ;

    return ;
}

static void get_f77(FILEVAR *fv)
/*++++++++++++++++
.PURPOSE  Edit the program for 1 file
.RETURNS  ---
.REMARKS  All details saved in last_var
-----------------*/
{
  FILINFO *pf ;
  ID *pid, *coo ;
  int fno ;
  char *vlist, *index;

    pf = fv->pf;
    pid = fv->pid ;
    coo = fv->coo ;
    fno = fv->fno ;
    index = oall ? "i__" : "" ;

    printf("\n%s\n", Crule) ;
    printf("\nC  Loading file '%s'\t! ", pf->dfile) ;
    prt(pf->text) ; putchar('\n') ;
    printf("\nC  Format for file interpretation\n\n") ;
    vlist = fv->fmt ;
    printf("%5d format(", fno) ;
    if ((int)strlen(vlist) < 57) printf("%s", vlist) ;
    else   prc(vlist);
    printf(")\n") ;

    vlist = fv->vlist ;
    printf("\nC  Effective file loading\n\n") ;
    printf("      open(unit=1,file='") ;
    if (ddat) printf("%s/", ddat) ;
    printf("%s', status='old')\n", pf->dfile) ;
    printf("      write(6,*) '....Loading file: %s'\n", pf->dfile) ;
    printf("      do i__=1,%d\n", pf->records);
    printf("        read(1,'(A%d)')%s\n", pf->reclen, fv->recname) ;
    printf("        read(%s,%d)", fv->recname, fno) ;
    if ((int)strlen(vlist) < 50) printf("%s", vlist) ;
    else   prc(vlist) ;
    putchar('\n') ;
    set_null(pf->tabform, pid, index, fv->recname) ;
    if (coo) set_coo(pf->tabform, pid, index, coo) ;
    printf("c    ..............Just test output...........\n") ;
    printf("        write(6,%d)", fno) ;
    if ((int)strlen(vlist) < 50) printf("%s", vlist) ;
    else   prc(vlist) ;
    putchar('\n') ;
    if (coo) {
	printf("        write(6,'(6H Pos: 2F8.4)') %s", coo->name) ;
	if (oall) printf("(%s)", index) ;
	coo = coo->next ;
	printf(",%s", coo->name); if (oall) printf("(%s)", index) ;
	coo = coo->prev ;
	putchar('\n') ;
    }
    printf("c    .......End.of.Just test output...........\n") ;
    printf("      end do\n") ;
    printf("      close(%d)\n", 1) ;
}

/*==================================================================
		Test Program
 *==================================================================*/

#define pVal(text)      \
	if (! *++p) {                           	\
	    if (--argc <= 0) error(text);   		\
	    else    p = *++argv;            		\
	}

int main (int argc, char **argv)
/*++++++++++++++++
.PURPOSE  Just a TEST program
-----------------*/
{
  CATINFO *tinfo; 
  FILINFO *fc, *fe, **af, **afiles ;
  FILEVAR *fv ;
  int fop, i;
  char *p, *getenv() ;

    /* Check the Arguments */
    fop = 0 ;
    while (-- argc > 0) {
	p = *++argv;
	if (*p != '-') break;	/* Options exhausted */
	switch (*++p) {
	  case 'd' :			/* Tell  the Directory 	*/
		pVal("Missing -d Directory");
		ddat = p; 
		continue;
	  case 'i' :			/* Tell which "ReadMe" 	*/
		pVal("Missing -i FileName");
		intro = p; i = strbloc(p, dirsep);
		if ((!ddat) && (i>=0)) {  /* Keep the path 	*/
		    ddat = strdup(intro);
		    ddat[i+1] = 0;
		}
		continue;
	  case 'a':
		oall = 1 ;
		continue ;
	  case 'v':                     /* Verbose      */
		if (strncmp(p, "ver", 3) == 0) {
		    printf("readme2f (CDS) Version %s\n", VERSION);
		    exit(0);
		}
		verbop = 1; fop |= FOP_v;
		/* osmtrace(1); */
		continue;
	  case 'h': 
		if (strcmp(p, "help") == 0) { 
		    printf("%s\n%s", usage, help); 
		    exit(0); 
		}
	  	goto bad_option;
	  default:
	  bad_option:
		fprintf(stderr, "\n****Bad option: -%s\n", p);
		exit(1) ;
	} 
    } 

    /* Header */
    printf("      program load_ReadMe\n"); 
    puts(Crule) ;
    printf("C  F77-compliant program generated by %s, on %s\n", 
	PGM, date('-')) ;
    puts(Crule) ;
    printf("%s", header) ;
    puts(Crule) ;
    putchar('\n') ;
    printf("      implicit none\n") ;
    printf("%s", explain_null) ;
    printf("      real*4     rNULL__\n");
    printf("      integer*4  iNULL__\n");
    printf("      parameter  (rNULL__=-1.e37)     \t! NULL real number\n");
    printf("      parameter  (iNULL__=-2147483647)\t! NULL int  number\n");
    
#if DEBUG
    printf("----%d arguments remain", argc);
    if (argc>0) printf(", first is %s\n", *argv);
#endif

#if 0
    if ((!ddat) && (!intro) && (argc>0)) {	/* File may be Dir/File */
	p = *argv;
	i = strbloc(p, dirsep);
	if (i >= 0) {
	    ddat = p; 
	    p += i;
	    *(p++) = 0;
	}
	intro = p;
	argc--; argv++;
    }
#else	/* V1.5 */
    if ((!ddat) && (!intro) && (argc==1)) {	/* File may be Dir/File */
	if (strchr(*argv, '/')) {
	    ddat = strdup(*argv);
	    if ((p = strstr(ddat, "/./"))) { *p = 0; *argv += 3+(p-ddat); }
	    else { p = strrchr(ddat, '/'); *p = 0; *argv += 1+(p-ddat); }
	}
    }
#endif

    if (ddat) {
    	if (chdir(ddat)) perror(ddat), 
    	    exit(1) ;
    }

    /* Set up the default variables */
    if (! intro) {
    	intro = "ReadMe";
    	if (!comfile(intro)) intro = "Intro";
    }

    if (verbop) fprintf(stderr, 
	"....call cad_new(\"%s\", 0x%X)\n", intro, fop);

    tinfo = cad_new(intro, fop);	/* Format File */
    if (!tinfo) exit(1) ;
    tinfo->ffile = intro;

    /* Write out the catalogue name + title */
    printf("\n%s\n", Crule) ;
    printf("Cat. ") ; 
    p = tinfo->section[0]->ebody ; i = *p ; *p = 0 ;
    prt(tinfo->section[0]->abody); /* fwrite(p, 1, strloc(p, '\n'), stdout) ; */
    *p = i ;
    printf("\n%s\n", Crule) ;
    if (verbop) cad_check(tinfo);

    printf("\nC  Internal variables\n\n") ;
    printf("      integer*4 i__\n") ;

    if (argc > 0) {
	af = afiles = (FILINFO **)malloc((argc+1)*sizeof(COLINFO *)) ;
	while (--argc >= 0) {
	    fc = cad_findfile(tinfo, *argv) ;
	    if (!fc) {
		fprintf(stderr, "****File '%s' not described\n", *argv) ;
		exit(1) ;
	    }
	    argv++ ;
	    *(af++) = fc ;
	}
	*af = (FILINFO *)0 ;
    }
    else {
	af = afiles = (FILINFO **)malloc((tinfo->nfinfo+1)*sizeof(COLINFO *)) ;
        for (fc=tinfo->afinfo, fe=fc+tinfo->nfinfo; fc<fe; fc++) {
	    if (!fc->tabform) continue ;
	    *(af++) = fc ;
	}
	*af = (FILINFO *)0 ;
    }

#if 0
    /* Calling the Subroutines */
    for (af = afiles; *af; af++) {
	fc = *af ;
	if (!fc->tabform) continue ;
	printf("      call load_%d\n", (af-afiles)) ;
    }
    printf("      stop\n") ;
    printf("      end\n") ;
#endif

    /* Writing out the Declarations */
    for (af = afiles; *af; af++) {
	fc = *af ;
	if (!fc->tabform) continue ;
	cad_f77(fc) ;
    }

    /* Reading the Data Files */
    if (!last_var) {
	fprintf(stderr, "****No data file****\n") ;
	exit(1) ;
    }
    for (fv = last_var; fv->prev; fv = fv->prev) ;
    for (; fv; fv = fv->next) get_f77(fv) ;

  /*close_f77:*/
    printf("\n%s\n", Crule) ;
    printf("      stop\n") ;
    printf("      end\n") ;
    
    cad_kill(tinfo);
    exit(0);
}
